@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2006-2007 Dan Pascu
+ * Copyright (C) 2006-2007 Dan Pascu. See LICENSE for details.
  * Author: Dan Pascu <dan@ag-projects.com>
  *
  * Fast JSON encoder/decoder implementation for Python
@@ -7,11 +7,11 @@
  */
 
 #include <Python.h>
+#include <stddef.h>
 #include <stdio.h>
 #include <ctype.h>
 #include <math.h>
 
-
 typedef struct JSONData {
     char *str; // the actual json string
     char *end; // pointer to the string end
@@ -19,7 +19,6 @@ typedef struct JSONData {
     int  all_unicode; // make all output strings unicode if true
 } JSONData;
 
-
 static PyObject* encode_object(PyObject *object);
 static PyObject* encode_string(PyObject *object);
 static PyObject* encode_unicode(PyObject *object);
@@ -42,19 +41,29 @@ static PyObject *JSON_EncodeError;
 static PyObject *JSON_DecodeError;
 
 
+#if PY_VERSION_HEX < 0x02050000
+typedef int Py_ssize_t;
+#define PY_SSIZE_T_MAX INT_MAX
+#define PY_SSIZE_T_MIN INT_MIN
+
+#define SSIZE_T_F "%d"
+#else
+#define SSIZE_T_F "%zd"
+#endif
+
 #define True  1
 #define False 0
 
 #ifndef INFINITY
-# define INFINITY HUGE_VAL
+#define INFINITY HUGE_VAL
 #endif
 
 #ifndef NAN
-# define NAN (HUGE_VAL - HUGE_VAL)
+#define NAN (HUGE_VAL - HUGE_VAL)
 #endif
 
 #ifndef Py_IS_NAN
-# define Py_IS_NAN(X) ((X) != (X))
+#define Py_IS_NAN(X) ((X) != (X))
 #endif
 
 #define skipSpaces(d) while(*((d)->ptr) && isspace(*((d)->ptr))) (d)->ptr++
@@ -65,7 +74,7 @@ static PyObject *JSON_DecodeError;
 static PyObject*
 decode_null(JSONData *jsondata)
 {
-    int left;
+    ptrdiff_t left;
 
     left = jsondata->end - jsondata->ptr;
 
@@ -84,7 +93,7 @@ decode_null(JSONData *jsondata)
 static PyObject*
 decode_bool(JSONData *jsondata)
 {
-    int left;
+    ptrdiff_t left;
 
     left = jsondata->end - jsondata->ptr;
 
@@ -108,7 +117,8 @@ static PyObject*
 decode_string(JSONData *jsondata)
 {
     PyObject *object;
-    int c, escaping, len, has_unicode, string_escape;
+    int c, escaping, has_unicode, string_escape;
+    Py_ssize_t len;
     char *ptr;
 
     // look for the closing quote
@@ -118,8 +128,8 @@ decode_string(JSONData *jsondata)
         c = *ptr;
         if (c == 0) {
             PyErr_Format(JSON_DecodeError,
-                         "unterminated string starting at position %d",
-                         jsondata->ptr - jsondata->str);
+                         "unterminated string starting at position " SSIZE_T_F,
+                         (Py_ssize_t)(jsondata->ptr - jsondata->str));
             return NULL;
         }
         if (!escaping) {
@@ -165,20 +175,20 @@ decode_string(JSONData *jsondata)
         PyErr_Fetch(&type, &value, &tb);
         if (type == NULL) {
             PyErr_Format(JSON_DecodeError,
-                         "invalid string starting at position %d",
-                         jsondata->ptr - jsondata->str);
+                         "invalid string starting at position " SSIZE_T_F,
+                         (Py_ssize_t)(jsondata->ptr - jsondata->str));
         } else {
             if (PyErr_GivenExceptionMatches(type, PyExc_UnicodeDecodeError)) {
                 reason = PyObject_GetAttrString(value, "reason");
                 PyErr_Format(JSON_DecodeError, "cannot decode string starting"
-                             " at position %d: %s",
-                             jsondata->ptr - jsondata->str,
+                             " at position " SSIZE_T_F ": %s",
+                             (Py_ssize_t)(jsondata->ptr - jsondata->str),
                              reason ? PyString_AsString(reason) : "bad format");
                 Py_XDECREF(reason);
             } else {
                 PyErr_Format(JSON_DecodeError,
-                             "invalid string starting at position %d",
-                             jsondata->ptr - jsondata->str);
+                             "invalid string starting at position " SSIZE_T_F,
+                             (Py_ssize_t)(jsondata->ptr - jsondata->str));
             }
         }
         Py_XDECREF(type);
@@ -196,7 +206,7 @@ static PyObject*
 decode_inf(JSONData *jsondata)
 {
     PyObject *object;
-    int left;
+    ptrdiff_t left;
 
     left = jsondata->end - jsondata->ptr;
 
@@ -224,7 +234,7 @@ static PyObject*
 decode_nan(JSONData *jsondata)
 {
     PyObject *object;
-    int left;
+    ptrdiff_t left;
 
     left = jsondata->end - jsondata->ptr;
 
@@ -295,9 +305,8 @@ decode_number(JSONData *jsondata)
     Py_DECREF(str);
 
     if (object == NULL) {
-        PyErr_Format(JSON_DecodeError, "invalid number starting at position %d",
-                     jsondata->ptr - jsondata->str);
-        puts(ptr);
+        PyErr_Format(JSON_DecodeError, "invalid number starting at position "
+                     SSIZE_T_F, (Py_ssize_t)(jsondata->ptr - jsondata->str));
     } else {
         jsondata->ptr = ptr;
     }
@@ -324,12 +333,14 @@ decode_array(JSONData *jsondata)
         c = *jsondata->ptr;
         if (c == 0) {
             PyErr_Format(JSON_DecodeError, "unterminated array starting at "
-                         "position %d", start - jsondata->str);
+                         "position " SSIZE_T_F,
+                         (Py_ssize_t)(start - jsondata->str));
             goto failure;;
         } else if (c == ']') {
             if (expect_item && items>0) {
                 PyErr_Format(JSON_DecodeError, "expecting array item at "
-                             "position %d", jsondata->ptr - jsondata->str);
+                             "position " SSIZE_T_F,
+                             (Py_ssize_t)(jsondata->ptr - jsondata->str));
                 goto failure;
             }
             jsondata->ptr++;
@@ -337,7 +348,8 @@ decode_array(JSONData *jsondata)
         } else if (c == ',') {
             if (expect_item) {
                 PyErr_Format(JSON_DecodeError, "expecting array item at "
-                             "position %d", jsondata->ptr - jsondata->str);
+                             "position " SSIZE_T_F,
+                             (Py_ssize_t)(jsondata->ptr - jsondata->str));
                 goto failure;
             }
             expect_item = True;
@@ -383,12 +395,14 @@ decode_object(JSONData *jsondata)
         c = *jsondata->ptr;
         if (c == 0) {
             PyErr_Format(JSON_DecodeError, "unterminated object starting at "
-                         "position %d", start - jsondata->str);
+                         "position " SSIZE_T_F,
+                         (Py_ssize_t)(start - jsondata->str));
             goto failure;;
         } else if (c == '}') {
             if (expect_key && items>0) {
                 PyErr_Format(JSON_DecodeError, "expecting object property name"
-                             " at position %d", jsondata->ptr - jsondata->str);
+                             " at position " SSIZE_T_F,
+                             (Py_ssize_t)(jsondata->ptr - jsondata->str));
                 goto failure;
             }
             jsondata->ptr++;
@@ -396,7 +410,8 @@ decode_object(JSONData *jsondata)
         } else if (c == ',') {
             if (expect_key) {
                 PyErr_Format(JSON_DecodeError, "expecting object property name"
-                             "at position %d", jsondata->ptr - jsondata->str);
+                             "at position " SSIZE_T_F,
+                             (Py_ssize_t)(jsondata->ptr - jsondata->str));
                 goto failure;
             }
             expect_key = True;
@@ -404,9 +419,9 @@ decode_object(JSONData *jsondata)
             continue;
         } else {
             if (c != '"') {
-                PyErr_Format(JSON_DecodeError,
-                             "expecting property name in object at "
-                             "position %d", jsondata->ptr - jsondata->str);
+                PyErr_Format(JSON_DecodeError, "expecting property name in "
+                             "object at position " SSIZE_T_F,
+                             (Py_ssize_t)(jsondata->ptr - jsondata->str));
                 goto failure;
             }
 
@@ -416,9 +431,9 @@ decode_object(JSONData *jsondata)
 
             skipSpaces(jsondata);
             if (*jsondata->ptr != ':') {
-                PyErr_Format(JSON_DecodeError,
-                             "missing colon after object property name at "
-                             "position %d", jsondata->ptr - jsondata->str);
+                PyErr_Format(JSON_DecodeError, "missing colon after object "
+                             "property name at position " SSIZE_T_F,
+                             (Py_ssize_t)(jsondata->ptr - jsondata->str));
                 Py_DECREF(key);
                 goto failure;
             } else {
@@ -527,16 +542,17 @@ encode_string(PyObject *string)
     size_t newsize = 2 + 6 * op->ob_size;
     PyObject *v;
 
-    if (newsize > INT_MAX) {
+    if (op->ob_size > (PY_SSIZE_T_MAX-2)/6) {
         PyErr_SetString(PyExc_OverflowError,
                         "string is too large to make repr");
+        return NULL;
     }
     v = PyString_FromStringAndSize((char *)NULL, newsize);
     if (v == NULL) {
         return NULL;
     }
     else {
-        register int i;
+        register Py_ssize_t i;
         register char c;
         register char *p;
         int quote;
@@ -593,7 +609,7 @@ encode_unicode(PyObject *unicode)
 {
     PyObject *repr;
     Py_UNICODE *s;
-    int size;
+    Py_ssize_t size;
     char *p;
 
     static const char *hexdigit = "0123456789abcdef";
@@ -601,6 +617,12 @@ encode_unicode(PyObject *unicode)
     s = PyUnicode_AS_UNICODE(unicode);
     size = PyUnicode_GET_SIZE(unicode);
 
+    if (size > (PY_SSIZE_T_MAX-2-1)/6) {
+        PyErr_SetString(PyExc_OverflowError,
+                        "unicode object is too large to make repr");
+        return NULL;
+    }
+
     repr = PyString_FromStringAndSize(NULL, 2 + 6*size + 1);
     if (repr == NULL)
         return NULL;
@@ -736,7 +758,7 @@ encode_unicode(PyObject *unicode)
 static PyObject*
 encode_tuple(PyObject *tuple)
 {
-    int i, n;
+    Py_ssize_t i, n;
     PyObject *s, *temp;
     PyObject *pieces, *result = NULL;
     PyTupleObject *v = (PyTupleObject*) tuple;
@@ -801,7 +823,7 @@ Done:
 static PyObject*
 encode_list(PyObject *list)
 {
-    int i;
+    Py_ssize_t i;
     PyObject *s, *temp;
     PyObject *pieces = NULL, *result = NULL;
     PyListObject *v = (PyListObject*) list;
@@ -884,7 +906,7 @@ Done:
 static PyObject*
 encode_dict(PyObject *dict)
 {
-    int i;
+    Py_ssize_t i;
     PyObject *s, *temp, *colon = NULL;
     PyObject *pieces = NULL, *result = NULL;
     PyObject *key, *value;
@@ -1058,18 +1080,20 @@ JSON_decode(PyObject *self, PyObject *args, PyObject *kwargs)
 
     object = decode_json(&jsondata);
 
-    Py_DECREF(str);
-
     if (object != NULL) {
         skipSpaces(&jsondata);
         if (jsondata.ptr < jsondata.end) {
-            Py_DECREF(object);
             PyErr_Format(JSON_DecodeError, "extra data after JSON description"
-                        " at position %d", jsondata.ptr-jsondata.str);
+                         " at position " SSIZE_T_F,
+                         (Py_ssize_t)(jsondata.ptr - jsondata.str));
+            Py_DECREF(str);
+            Py_DECREF(object);
             return NULL;
         }
     }
 
+    Py_DECREF(str);
+
     return object;
 }
 
